#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes to serialize and deserialize messages.

.. autosummary::
    :toctree: _codecs

    wotpy.codecs.base
    wotpy.codecs.enums
    wotpy.codecs.json_codec
    wotpy.codecs.text
"""
