#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Implementations of the supported Protocol Binding templates.

.. autosummary::
    :toctree: _protocols

    wotpy.protocols.coap
    wotpy.protocols.http
    wotpy.protocols.mqtt
    wotpy.protocols.ws
    wotpy.protocols.client
    wotpy.protocols.enums
    wotpy.protocols.exceptions
    wotpy.protocols.server
    wotpy.protocols.utils
"""
