#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
CoAP resources that implement each of the Interaction verbs.

.. autosummary::
    :toctree: _resources

    wotpy.protocols.coap.resources.action
    wotpy.protocols.coap.resources.event
    wotpy.protocols.coap.resources.property
    wotpy.protocols.coap.resources.utils
"""
