#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
ConsumedThing and related entities.

.. autosummary::
    :toctree: _consumed

    wotpy.wot.consumed.interaction_map
    wotpy.wot.consumed.thing
"""
