#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Constants related to objects in the Thing hierarchy.
"""

WOT_TD_CONTEXT_URL = "https://w3c.github.io/wot/w3c-wot-td-context.jsonld"
"""W3C WoT TD semantic context."""

WOT_COMMON_CONTEXT_URL = "https://w3c.github.io/wot/w3c-wot-common-context.jsonld"
"""W3C WoT common semantic context."""
